#Script for Future Compass Paper Analysis
#As of: 2023-03-28
#Author: Hiroaki Sugino

#----- ///// Introduction Part ///// -----

#----- Background -----

#This script has been written for the analysis on the data set about "future compass."
#This research is conducted by Sugino, H., Sekiguchi, T., Terada, Y. and Hayashi, N.
#The objective of this analysis and paper is to visualize residents' future vision for Tokyo Bay with vectors in a correlation circle.

#+++ Major Analysis Procedure +++
#1) Descriptive analysis on the distribution of respondents
#2) BTM (Biterm Topic Modelling)
#3) MFA (Multiple Factor Analysis) and vector visualization of future vision


#----- Power References -----

##### <Biterm Topic Modeling>
#Yan et al. (2013) A Biterm Topic model for Short Texts
#Yan, Xiaohui & Guo, Jiafeng & Lan, Yanyan & Cheng, Xueqi. (2013). A biterm topic model for short texts. WWW 2013 - Proceedings of the 22nd International Conference on World Wide Web. 1445-1456. 10.1145/2488388.2488514. 

#About {BTM} package: Biterm Topic Models for Short Text
#https://cran.r-project.org/web/packages/BTM/index.html

#Biterm topic modelling for short texts
#https://www.r-bloggers.com/2020/04/biterm-topic-modelling-for-short-texts/

#Text Mining with R
#https://www.tidytextmining.com/index.html


##### <Multi-factor Analysis>
#Escofier, B. and Pages, J. (1994) Multiple Factor Analysis (AFMULT package). Computational Statistics and Data Analysis, 18, 121-140. 
#https://doi.org/10.1016/0167-9473(94)90135-X

#Becue-Bertaut, M. and Pages, J. (2008) Multiple factor analysis and clustering of a mixture of quantitative, categorical and frequency data. Computational Statistice and Data Analysis, 52, 32553268.
#https://doi.org/10.1016/j.csda.2007.09.023

#Husson, Francois, Julie Josse, Sebastien Le, and Jeremy Mazet. 2017. FactoMineR: Multivariate Exploratory Data Analysis and Data Mining. 
#https://CRAN.R-project.org/package=FactoMineR.

#Kassambara, Alboukadel, and Fabian Mundt. 2017. Factoextra: Extract and Visualize the Results of Multivariate Data Analyses. 
#http://www.sthda.com/english/rpkgs/factoextra.


#----- ///// Analysis Part ///// -----

#----- Environment Setting -----

#Clear the console, plots, and environment values if needed
rm(list = ls(all = TRUE))   #Clear the values and dataframes in the environment
if(dev.cur() > 1) dev.off() #Clear the plots and plot-related-setting
cat("\014")                 #Clear the console

#Check the current working directory
getwd()
#setwd("C:/Users/hsugi/GWD/z0_RWD/future_compass")　#If the path to the working directory is not what you meant, run this line deleting # in the beginning and designate your intended path instead of the dummy
myworkingdirectry <- getwd() #Store the path to the working directory
list.files() #Check the files in the current working directory

#Define the information to use when writing out files
outinfo1 <- "fc"       #Define the prefix code for the files written out in this project
outinfo2 <- Sys.Date()   #Define the date of work

#In order to ensure reproducibility, set a seed of random numbers
set.seed(1234567)


#----- Package Preparation -----

#Make a list of required packages

lrp <- c(
  #General
  "RColorBrewer", #Beautiful color palette
  
  #Data handling
  "tidyverse",    #A set of packages for data handling(dplyr, tibble, ggplot2, forcats, readr, stringr, tidyr, purrr)
  "reshape2",     #For data reshaping
  "magrittr",     #For convenient data handling
  "tableone",     #To make descriptive table for factor type data
  "skimr",        #To skim the dataset
  
  #Data analysis
  "psych",        #For many useful functions of psycho-analysis(e.g. "describe" for descriptive analysis)
  "GPArotation",  #For some rotation methods of factor analysis
  "corrplot",     #For visualization of correlation matrix
  "Rgraphviz",    #For network visualization
  "igraph",       #For network analysis
  "factoextra",   #For multi-factor analysis
  "FactoMineR",   #For multi-factor analysis
  "tidytext",     #For convenient and comfortable text vector processing
  "text2map",     #To make a document-term matrix fast, tools for text matrices, embeddings, and networks
  
  #Data analysis specialized fro biterm topic modeling
  "BTM",          #Base package for bi-term topic modeling
  "udpipe",       #This natural language processing toolkit 
  "textplot",     #For visualization of complex relations in texts
  "concaveman",   #To compute the concave polygon(s) for one or several set of points
  "stopwords",    #Supplementary package to conduct BTM
  "data.table",   #Supplementary package to conduct BTM

  #Data visualization
  "ggplot2",      #For figure visualization
  "ggsci",        #For additional colorful palette used in ggplot2
  "ggpubr",       #For making publication-ready figures (e.g. GG boxplot)
  "GGally",       #Extension to "ggplot2" to draw good-looking figures (e.g. pairwise plots)
  "ggdist",       #For visualizing distributions
  "gtsummary",    #To make publication ready summary tables
  "tidyquant"     #To use quantitative functions with the 'tidyverse' functions 
)

#Install & library the required packages if not installed or library the required packages already installed
for (i in 1:length(lrp)) {
  ptp <- lrp[i]
  if(!require(ptp, character.only = TRUE)) {
    install.packages(ptp, dependencies = TRUE)
    library(ptp, character.only = TRUE); rm(ptp)
  } else {
    library(ptp, character.only = TRUE); rm(ptp)
  }
}

#If the script above doesn't work well, please download/library required package(s) manually:
#install.packages("psych", dependencies = TRUE)
#library(psych)


#----- 1) Data Import, Processing and Descriptive Analysis -----

#----- Data import and demographic statistics -----
d <- read.csv("data4fc_sugino_etal.csv", header = TRUE, row.names = 1, fileEncoding = "cp932")

#Make the followings into factor variables
d$pref <- recode_factor(d$pref, "Kanagawa" = "Kanagawa", "Tokyo" = "Tokyo", "Chiba" = "Chiba")
d$gender <- recode_factor(d$gender,  "Male" = "Male", "Female" = "Female",)
d$age_c <- recode_factor(d$age_c, "AC2" = "AC2", "AC3" = "AC3", "AC4" = "AC4", "AC5" = "AC5", "AC6" = "AC6")

#Data check
str(d) 　　
skim(d)
summary(d)
describe(d)

#Summary table for gender and prefecture
d %>%
  select(pref, gender) %>%
  tbl_summary(by = pref) %>%
  add_overall %>%
  as_flex_table() 
#%>%
# flextable::save_as_docx(path = "gender_pref_table.docx")

#Summarize age data
summary(d$age)  
psych::describe(d$age)  

write.csv(
  describe(d$age), str_glue("{outinfo1}-describe_age-{outinfo2}.csv")
)

#----- Visualize distributions of age data for each gender and prefecture category -----

#Half-density plot / Raincloud plot
#The boxplot does not show where densities are clustered, but the raincloud plot does.

gp_dist <- d %>%
  select(pref, gender, age) %>%
  ggplot(aes(y = age, fill = gender)) +
  
  # add half-violin from {ggdist} package
  ggdist::stat_halfeye(
    ## custom bandwidth
    adjust = 0.4,
    ## move geom to the right
    justification = -.25,
    ## remove slab interval
    .width = 0,
    point_colour = NA,
    alpha = 0.6
  ) +
  # Add box plot
  geom_boxplot(
    width = .3,
    ## remove outliers
    outlier.color = NA,
    alpha = 0.6
  ) +
  # Add dot plots from {ggdist} package
  ggdist::stat_dots(
    ## orientation to the left
    side = "left",
    ## move geom to the left
    justification = 1.25,
    ## adjust grouping (binning) of observations
    binwidth = 0.8,
    alpha = 0.6
  ) +
  #Adjust theme
  scale_fill_tq() + 
  theme_tq(base_size = 16) +
  labs(
    x = NULL,
    y = "Age",
    fill = "Gender"
  ) +
  facet_grid(~ pref) +
  guides(x = "none") +
  ggpubr::theme_pubclean() + 
  theme(legend.position = "bottom")

gp_dist


#----- 2) Bi-term Topic Modelling & Structural Topic Modeling -----

library(udpipe)
library(data.table)
library(stopwords)

#----- Data preparation -----
d4btm_nfv <- d %>% dplyr::select(-q5t_en) 
d4btm_pfv <- d %>% dplyr::select(-q4t_en)

d4btm_nfv$text <- tolower(d4btm_nfv$q4t_en)
d4btm_nfv$text <- gsub("none in particular", "", d4btm_nfv$text)
d4btm_nfv$text <- gsub("'", "", d4btm_nfv$text)
d4btm_nfv$text <- gsub("<.+>", "", d4btm_nfv$text) %>% enc2utf8()

d4btm_pfv$text <- tolower(d4btm_pfv$q5t_en)
d4btm_pfv$text <- gsub("none in particular", "", d4btm_pfv$text)
d4btm_pfv$text <- gsub("'", "", d4btm_pfv$text)
d4btm_pfv$text <- gsub("<.+>", "", d4btm_pfv$text) %>% enc2utf8()


#Tokenising, Lemmatising, Tagging and Dependency Parsing of raw text in TIF format
anno_nfv <- udpipe(d4btm_nfv, "english", trace = 10)
anno_pfv <- udpipe(d4btm_pfv, "english", trace = 10)

#Extracting cooccurrences of nouns, adjectives and verbs within 3 words distance
biterms_nfv <- as.data.table(anno_nfv)
biterms_nfv <- biterms_nfv[, cooccurrence(x = lemma,
                                          relevant = upos %in% c("NOUN", "ADJ", "VERB") & 
                                            nchar(lemma) > 2 & !lemma %in% stopwords("en"),
                                          skipgram = 3),
                           by = list(doc_id)]

biterms_pfv <- as.data.table(anno_pfv)
biterms_pfv <- biterms_pfv[, cooccurrence(x = lemma,
                                          relevant = upos %in% c("NOUN", "ADJ", "VERB") & 
                                            nchar(lemma) > 2 & !lemma %in% stopwords("en"),
                                          skipgram = 3),
                           by = list(doc_id)]

#Data preparation for BTM
traindata_nfv <- subset(anno_nfv, upos %in% c("NOUN", "ADJ", "VERB") & !lemma %in% stopwords("en") & nchar(lemma) > 2)
traindata_nfv <- traindata_nfv[, c("doc_id", "lemma")]

traindata_pfv <- subset(anno_pfv, upos %in% c("NOUN", "ADJ", "VERB") & !lemma %in% stopwords("en") & nchar(lemma) > 2)
traindata_pfv <- traindata_pfv[, c("doc_id", "lemma")]


#----- Check appropriate number of topics and build BTM model -----

library(BTM)
set.seed(1234567)

#modified function, originally calculating semantic coherence in LDA models provided by @grenwi, provided by @abitter
#https://github.com/bnosac/BTM/issues/3#issuecomment-742515859
#
#modified from tmca_coherence() function (by A. Niekler & G. Wiedemann)
#https://github.com/tm4ss/tm4ss.github.io
coherenceBTM <- function(model, DTM, N = 10) {
  
  # Ensure matrix or Matrix-format (convert if slam)
  require(Matrix)
  require(slam)
  if (is.simple_triplet_matrix(DTM)) {
    DTM <- sparseMatrix(i=DTM$i, j=DTM$j, x=DTM$v, dims=c(DTM$nrow, DTM$ncol), dimnames = dimnames(DTM))
  }
  
  K <- model$K
  
  DTMBIN <- DTM > 0
  
  documentFrequency <- colSums(DTMBIN)
  names(documentFrequency) <- colnames(DTMBIN)
  
  topNtermsPerTopic <- terms(model, top_n = N)
  
  termcollect <- list()
  for (i in 1:K){
    termcollect[[i]] <- topNtermsPerTopic[[i]][,1]
  }
  
  allTopicModelTerms <- unique(as.vector(unlist(termcollect)))
  
  DTMBIN <- DTMBIN[, allTopicModelTerms]
  DTMBINCooc <- t(DTMBIN) %*% DTMBIN
  DTMBINCooc <- t((DTMBINCooc + 1) / colSums(DTMBIN))
  DTMBINCooc <- log(DTMBINCooc)
  DTMBINCooc <- as.matrix(DTMBINCooc)
  
  coherence <- rep(0, K)
  pb <- txtProgressBar(max = K)
  for (topicIdx in 1:K) {
    setTxtProgressBar(pb, topicIdx)
    topWordsOfTopic <- topNtermsPerTopic[[topicIdx]][,1]
    
    coherence[topicIdx] <- 0
    for (m in 2:length(topWordsOfTopic)) {
      for (l in 1:(m-1)) {
        mTerm <- as.character(topWordsOfTopic[m])
        lTerm <- as.character(topWordsOfTopic[l])
        coherence[topicIdx] <- coherence[topicIdx] + DTMBINCooc[mTerm, lTerm]
      }
    }
  }
  close(pb)
  
  return(coherence)
}

#Function to search appropriate K for BTM

searchK4btm <- function(traindata4btm, bitermlist4btm, K, iter_n, boot_n, suffix) {
  
  require(text2map)
  
  ll_list <- NULL
  cs_list <- NULL
  
  traindata4btm_odpr <- aggregate(traindata4btm$lemma, by = list(traindata4btm$doc_id), paste, collapse = " ") #one doc per row
  names(traindata4btm_odpr) <- c("doc_id", "lemma")
  
  temp_index_df_all <- NULL
  
  for (j in 1:boot_n) {
    
    if(j==1) cat(paste("Bootstrap Milestone: 0 - 25 % ...\n", sep = ""))
    if(j==round(boot_n/4)) cat(paste("Bootstrap Milestone: 25 - 50 % ....\n", sep = ""))
    if(j==round(boot_n/4)*2) cat(paste("Bootstrap Milestone: 50 - 75 % ....\n", sep = ""))
    if(j==round(boot_n/4)*3) cat(paste("Bootstrap Milestone: 75 - 100 % ...\n", sep = ""))
    
    for (k in 1:length(K)) {
      temp_k <- K[k]
      temp_model <- BTM(traindata4btm, biterms = bitermlist4btm, k = temp_k, iter = iter_n, background = FALSE)
      temp_fit <- logLik(temp_model)
      temp_ll <- temp_fit$ll
      ll_list <- c(ll_list, temp_ll)
      
      DTM <- dtm_builder(
        traindata4btm_odpr,
        text = "lemma",
        doc_id = "doc_id"
      )
      
      temp_mean_cs <- mean(coherenceBTM(temp_model, DTM, N = 10)) # N is the number of top terms per topic to consider
      cs_list <- c(cs_list, temp_mean_cs)
    }
    
    temp_index_df <- data.frame(K, ll_list, cs_list)
    colnames(temp_index_df) <- c("N_Topics_k", "Log_Likelihood", "Coherence_Score")
    
    temp_index_df_all <- dplyr::bind_rows(temp_index_df_all, temp_index_df)
    
  }
  
  index_df <- temp_index_df_all %>% group_by(N_Topics_k) %>%
    dplyr::summarize(Log_Likelihood = mean(Log_Likelihood), Coherence_Score = mean(Coherence_Score))
  
  print(index_df)
  
  write.csv(index_df, 
            str_glue("index_df-{Sys.Date()}-{suffix}.csv")
  )
  
  #Plot
  par(oma = c(0, 0, 0, 2))
  
  plot(index_df$N_Topics_k, index_df$Log_Likelihood, 
       xlab = "Number of Topics (k)", ylab = "Log Likelihood",
       type = "b", pch = 15, cex = 2, lwd = 2, lty = 1, axes = FALSE)
  axis(1)
  axis(2)
  par(new = TRUE)
  plot(index_df$N_Topics_k, index_df$Coherence_Score, 
       xlab = "", ylab = "",
       type = "b", pch = 15, cex = 2, lwd = 2, lty = 2, axes = FALSE)
  mtext("Coherence Score", side = 4, line = 3) #right y label
  axis(4)
  
  par(oma = c(0, 0, 0 ,0))
  
  #box()
  legend("topleft", legend = c("Log Likelihood", "Coherence Score"), lty = c(1, 2))
  
}

#Display the results in the console, and as a plot. Excel file will be written down
searchK4btm(traindata_nfv, biterms_nfv, K = c(1:10), iter_n = 2000, boot_n = 10, suffix = "nfv") 
searchK4btm(traindata_pfv, biterms_pfv, K = c(1:10), iter_n = 2000, boot_n = 10, suffix = "pfv") 

k4nfv <- 8
k4pfv <- 6

#Build the biterm topic model with certain topics and provide the set of biterms to cluster upon
model_nfv <- BTM(traindata_nfv, biterms = biterms_nfv, k = k4nfv, iter = 2000, background = FALSE, trace = 100)
model_pfv <- BTM(traindata_pfv, biterms = biterms_pfv, k = k4pfv, iter = 2000, background = FALSE, trace = 100)


#----- Visualize the biterm topic clusters in a topic map -----

#Using the {textplot} package available at https://github.com/bnosac/textplot.
library(textplot)
library(ggraph)

#Without labels
plot(model_nfv, top_n = 10,
     title = "BTM model", subtitle = "Nagative Future Vision"
)

plot(model_pfv, top_n = 10, 
     title = "BTM model", subtitle = "Positive Future Vision"
)

#With labels

topic_labels_nfv <- c(
  "1. Uninhabitable Environment for Fish", 
  "2. External Influence from Human Society", 
  "3. Water Pollution and Environmental Destruction",
  "4. Vessel Accident and Negative Consequence",
  "5. Return to the Dirty Sea with Sludge", 
  "6. Water Quality Deterioration with Red/Blue Tide", 
  "7. Disaster such as Earthquake and Tsunami", 
  "8. Increase and Invasion of Alien Species"
)

topic_labels_pfv <- c(
  "1. Foundation for Cultural / Recreational Activities", 
  "2. Revival of Edomae(in front of Tokyo) Fish", 
  "3. Habitable Environment for Fish and Other Creatures",
  "4. Attractive Bay Close to the Tokyo Metropolitan Area",
  "5. Clean and Beautiful Sea / Coastal Zone", 
  "6. Improvement and Maintenance of Water Quality"
)

plot(model_nfv, top_n = 10,
     title = "BTM Topic Map - Nagative Future Vision",
     labels = topic_labels_nfv,
     which = c(1:k4nfv)
)

plot(model_pfv, top_n = 10,
     title = "BTM Topic Map - Positive Future Vision",
     labels = topic_labels_pfv,
     which = c(1:k4pfv)
)


#----- Get the list of tokens with highest probabilities for each topic -----

td_beta_nfv <- terms(model_nfv, type = c("tokens"), threshold = 0, top_n = 10)
td_beta_pfv <- terms(model_pfv, type = c("tokens"), threshold = 0, top_n = 10)

term_prob_vis <- function(td_beta) {
  
  td_beta_df <- lapply(td_beta, as.data.frame) %>%
    bind_rows(.id = "Topic") 
  
  td_beta_df$Topic <- as.integer(td_beta_df$Topic)
  
  options(repr.plot.width=7, repr.plot.height=8, repr.plot.res=100) 
  
  ggplot(td_beta_df, aes(x = reorder_within(token, probability, Topic), y = probability, fill = as.factor(Topic))) +
    geom_col(alpha = 0.8, show.legend = FALSE) +
    facet_wrap(~Topic, scales = "free_y") +
    coord_flip() +
    scale_x_reordered() +
    labs(x = NULL, y = expression(beta),
         title = "Top words with high probabilities for each topic"
    )
  
}

term_prob_vis(td_beta_nfv)
term_prob_vis(td_beta_pfv)

#Write down the data frame of term probabilities for each topic
td_beta_nfv_df <- lapply(td_beta_nfv, as.data.frame) %>%
  bind_rows(.id = "Topic") 
td_beta_pfv_df <- lapply(td_beta_pfv, as.data.frame) %>%
  bind_rows(.id = "Topic") 

write.csv(
  td_beta_nfv_df, str_glue("{outinfo1}-term_probabilities_nfv-{outinfo2}.csv")
)

write.csv(
  td_beta_pfv_df, str_glue("{outinfo1}-term_probabilities_pfv-{outinfo2}.csv")
)


#----- Obtain doc-topic matrix -----

docsize_nfv <- table(traindata_nfv$doc_id)
scores_nfv <- predict(model_nfv, traindata_nfv) %>% data.frame()
colnames(scores_nfv) <- rep(paste("Topic_nfv_", seq(1, k4nfv, by = 1), sep = "")) 
scores_nfv <- (round(scores_nfv, 2)) %>% data.frame()
scores_nfv$doc_id <- row.names(scores_nfv)

docsize_pfv <- table(traindata_pfv$doc_id)
scores_pfv <- predict(model_pfv, traindata_pfv) %>% data.frame()
colnames(scores_pfv) <- rep(paste("Topic_pfv_", seq(1, k4pfv, by = 1), sep = ""))
scores_pfv <- (round(scores_pfv, 2)) %>% data.frame()
scores_pfv$doc_id <- row.names(scores_pfv)

scores_all <- left_join(d, scores_nfv, by = "doc_id") %>% 
  left_join(., scores_pfv, by = "doc_id")

scores_all_selected <- scores_all %>% select(-c(doc_id, age, q4t_en, q5t_en))

write.csv(
  scores_all_selected, str_glue("{outinfo1}-scores_all-{outinfo2}.csv")
)


#----- 3) Multi-factor Analysis and vector visualization of future vision -----

colnames(scores_all_selected) <- c("Prefecture", "Gender", "Age_C", topic_labels_nfv, topic_labels_pfv)

res.mfa <- MFA(scores_all_selected,
               group = c(1, 1, 1, k4nfv, k4pfv),
               type = c("n", "n", "n", "s", "s"),
               name.group = c("Prefecture", "Gender", "Age_C", "Negative", "Positive")
)

fviz_mfa_var(res.mfa, "quanti.var", palette = "aaas", 
             col.var.sup = "violet", repel = TRUE)

fviz_mfa_var(res.mfa, "quanti.var", palette = "aaas", 
             col.var.sup = "violet", repel = TRUE, geom = c("arrow"))

# Contributions to dimension 1
fviz_contrib(res.mfa, choice = "quanti.var", axes = 1, top = 20,
             palette = "aaas") + theme(plot.margin = margin(0, 0, 0.5, 4, "cm"),
                                       axis.text.x = element_text(size = 12))

# Contributions to dimension 2
fviz_contrib(res.mfa, choice = "quanti.var", axes = 2, top = 20,
             palette = "aaas") + theme(plot.margin = margin(0, 0, 0.5, 4, "cm"),
                                       axis.text.x = element_text(size = 12))

#Graph of individuals
ind <- get_mfa_ind(res.mfa)
ind

fviz_mfa_ind(res.mfa, 
             habillage = "Prefecture", # color by groups 
             palette = c("#00AFBB", "#E7B800", "#FC4E07"),
             addEllipses = TRUE, ellipse.type = "confidence", 
             repel = TRUE # Avoid text overlapping
) 

fviz_mfa_ind(res.mfa, partial = "all")

fviz_ellipses(res.mfa, c("Prefecture", "Gender", "Age_C"), repel = TRUE)

#----- Script ends here -----

